/* tclif.c - df3tcl */

#include "df3tcl.h"

#define AUU __attribute__((unused))

/* use 'Tcl_Panic' */
#define APPPANIC appTclPanic(__FILE__,__LINE__)

#define MAXCBB 200

#define TOKEN "df3"

static char const * const cbmask = "%s %zu {%s} %s";

static char const * const nscreate = "namespace eval ::" TOKEN " {}";

static char const * const nslvmask = "::" TOKEN "::lv%zu";

static char const * const oemesg = "option ?arg ...?";

static char const * const smask = TOKEN " create -r %d %s";

static df3Coord const origin = {0, 0, 0};

/* ------------------------------------------------------------------------- */

static bool valueInRange(size_t const bpv, long const v) {
  long const cmp = 4 == bpv ? UINT32_MAX : 2 == bpv ? UINT16_MAX : UINT8_MAX;
  return 0 <= v  &&  v <= cmp;
}

static bool filePermCheck(char const * const fn, int const mode) {
  Tcl_Obj *tmp = Tcl_NewStringObj(fn, -1);
  Tcl_IncrRefCount(tmp);
  int const res = Tcl_FSAccess(tmp, mode);
  Tcl_DecrRefCount(tmp);
  return (bool)res;
}

static bool cornerCheck(Tcl_Interp *itp,
                        df3Axes const axes,
                        df3Coord const llc,
                        df3Coord const trc) {
  if (trc.x < llc.x  ||  trc.y < llc.y  ||  trc.z < llc.z) {
    Tcl_SetErrorCode(itp, "oops, TRC smaller than LLC error", NULL);
    return false;
  }
  else if (axes.x <= trc.x  ||  axes.y <= trc.y  ||  axes.z <= trc.z) {
    Tcl_SetErrorCode(itp, "oops, TRC outside volume error", NULL);
    return false;
  }
  return true;
}

static int list2coord(Tcl_Interp *itp, df3Coord *at, Tcl_Obj * const arg) {
  int n = 0;
  Tcl_ListObjLength(itp, arg, &n);
  if (3 != n) return TCL_ERROR;

  int v[3] = {0};
  for (size_t i = 0; 3 != i; ++i) {
    Tcl_Obj *tmp;
    Tcl_ListObjIndex(itp, arg, (int)i, &tmp);
    if (TCL_OK != Tcl_GetIntFromObj(itp, tmp, &v[i]))
      return TCL_ERROR;
    else if (0 > v[i]  ||  UINT16_MAX <= v[i])
      return TCL_ERROR;
  }

  at->x = (size_t)v[0], at->y = (size_t)v[1], at->z = (size_t)v[2];

  return TCL_OK;
}

static Tcl_Obj *coord2list(df3Coord const at) {
  Tcl_Obj *tmp[3];
  tmp[0] = Tcl_NewIntObj((int)at.x);
  tmp[1] = Tcl_NewIntObj((int)at.y);
  tmp[2] = Tcl_NewIntObj((int)at.z);
  Tcl_Obj *a = Tcl_NewListObj(3, tmp);
  Tcl_IncrRefCount(a);
  return a;
}

static void volumeFinish(Tcl_Interp *itp, Volume *vp) {
  char lnk[40] = {CHRNUL};
  df3Error const ec = df3_close_file(&vp->df3);
  if (ecOk != ec)
    if (app_diag)
      fprintf(stderr, "[II] error closing " TOKEN ".%zu: '%s'.\n",
              vp->id, df3_handle_error(&vp->df3, false));
  free(vp->df3.fname);
  sprintf(lnk, nslvmask, vp->id);
  Tcl_UnlinkVar(itp, lnk);
  Tcl_Free((char *)vp);
  return;
}

static df3Error volumeStartCreate(Volume *vp) {
  return df3_create_file(&vp->df3);
}

static df3Error volumeStartOpen(Volume *vp, bool const rdonly) {
  return df3_open_file(&vp->df3, rdonly);
}

/* ------------------------------------------------------------------------- */

static int tokenCmdAxes(Tcl_Interp *itp, Volume *vp) {
  Tcl_Obj *a = coord2list(vp->df3.axes);
  Tcl_SetObjResult(itp, a);
  Tcl_DecrRefCount(a);
  return TCL_OK;
}

static int tokenCmdBpv(Tcl_Interp *itp, Volume *vp) {
  Tcl_SetObjResult(itp, Tcl_NewIntObj((int)vp->df3.bpv));
  return TCL_OK;
}

static int tokenCmdForeach(Tcl_Interp *itp, Volume *vp, Tcl_Obj * const cb) {
  bool flag = false;
  char const * const proc = Tcl_GetString(cb);
  char lnk[40] = {CHRNUL},
       s[MAXCBB];
  df3Error ec;
  Tcl_Obj *tmp;

  sprintf(lnk, nslvmask, vp->id);
  for (size_t z = 0; z != vp->df3.axes.z; ++z)
    for (size_t y = 0; y != vp->df3.axes.y; ++y)
      for (size_t x = 0; x != vp->df3.axes.x; ++x) {
        df3Coord const at = {x, y, z};
        uint32_t rval = 0;
        if (ecOk != (ec = df3_get_voxel(&vp->df3, at, &rval))) {
          Tcl_SetErrorCode(itp, "oops, df3 read error", NULL);
          goto label_skip;
        }
        vp->lv = rval;
        memset(s, 0, MAXCBB);
        tmp = coord2list(at);
        sprintf(s, cbmask, proc, vp->df3.bpv, Tcl_GetString(tmp), lnk);
        Tcl_DecrRefCount(tmp);
        if (TCL_OK != Tcl_EvalEx(itp, s, -1, TCL_EVAL_GLOBAL))
          return TCL_ERROR;
        int rtn = 0;
        Tcl_GetBooleanFromObj(itp, Tcl_GetObjResult(itp), &rtn);
        if (!rtn) {
          /* TODO "break" as result? */
          flag = true;
          goto label_skip;
        }
        if (vp->lv != rval) {
          if (vp->df3.readonly) {
            Tcl_SetErrorCode(itp, "oops, df3 open readonly", NULL);
            goto label_skip;
          }
          if (ecOk != df3_put_voxel(&vp->df3, at, vp->lv)) {
            Tcl_SetErrorCode(itp, "oops, df3 write error", NULL);
            goto label_skip;
          }
        }
      }
  flag = true;

label_skip:
  if (flag) Tcl_ResetResult(itp);
  return flag ? TCL_OK : TCL_ERROR;
}

static int tokenCmdGet(Tcl_Interp *itp,
                       Volume *vp,
                       int ac,
                       Tcl_Obj * const av[]) {
  Tcl_Obj *list = Tcl_NewListObj(0, NULL);
  Tcl_IncrRefCount(list);

  for (size_t i = 0; i != (size_t const)ac; ++i) {
    df3Coord at = {0, 0, 0};
    if (TCL_OK != list2coord(itp, &at, av[i])) {
      Tcl_SetErrorCode(itp, "oops, bad coordinate '",
              Tcl_GetString(av[i]), "'.", NULL);
      return TCL_ERROR;
    }
    else if (!cornerCheck(itp, vp->df3.axes, origin, at)) {
      Tcl_SetErrorCode(itp, "oops, coord outside volume", NULL);
      return TCL_ERROR;
    }
    uint32_t v = 0;
    if (ecOk != df3_get_voxel(&vp->df3, at, &v)) {
      if (app_diag)
        fprintf(stderr, "[II] df3 read error: '%s'.\n",
                df3_handle_error(&vp->df3, false));
      Tcl_SetErrorCode(itp, "oops, df3 read error", NULL);
      return TCL_ERROR;
    }
    Tcl_ListObjAppendElement(itp, list, av[i]);
    Tcl_ListObjAppendElement(itp, list, Tcl_NewLongObj((long)v));
  }

  Tcl_SetObjResult(itp, list);
  Tcl_DecrRefCount(list);
  return TCL_OK;
}


static int tokenCmdInfo(Tcl_Interp *itp, Volume *vp) {
  Tcl_Obj *tmp[10];
  tmp[0] = Tcl_NewStringObj("file", -1);
  tmp[1] = Tcl_NewStringObj(vp->df3.fname, -1);
  tmp[2] = Tcl_NewStringObj("axes", -1);
  tmp[3] = coord2list(vp->df3.axes);
  tmp[4] = Tcl_NewStringObj("bpv", -1);
  tmp[5] = Tcl_NewIntObj((int)vp->df3.bpv);
  tmp[6] = Tcl_NewStringObj("readonly", -1);
  tmp[7] = Tcl_NewBooleanObj(vp->df3.readonly);
  tmp[8] = Tcl_NewStringObj("writes", -1);
  tmp[9] = Tcl_NewLongObj((long)vp->df3.nwrite);
  Tcl_SetObjResult(itp, Tcl_NewListObj(10, tmp));
  Tcl_DecrRefCount(tmp[3]);
  return TCL_OK;
}


static int tokenCmdPut(Tcl_Interp *itp,
                       Volume *vp,
                       int ac,
                       Tcl_Obj * const av[]) {
  if (vp->df3.readonly) {
    Tcl_SetErrorCode(itp, "oops, cannot modify 'readonly' volume.", NULL);
    return TCL_ERROR;
  }

  for (size_t i = 0; i < (size_t const)ac; i += 2) {
    df3Coord at = {0, 0, 0};
    long tmp = 0;
    if (TCL_OK != list2coord(itp, &at, av[i])) {
      Tcl_SetErrorCode(itp, "oops, bad coordinate '",
              Tcl_GetString(av[i]), "'.", NULL);
      return TCL_ERROR;
    }
    else if (!cornerCheck(itp, vp->df3.axes, origin, at)) {
      Tcl_SetErrorCode(itp, "oops, coord outside volume", NULL);
      return TCL_ERROR;
    }
    else if (TCL_OK != Tcl_GetLongFromObj(itp, av[i + 1], &tmp)) {
      Tcl_SetErrorCode(itp, "oops, bad value '",
              Tcl_GetString(av[i + 1]), "'.", NULL);
      return TCL_ERROR;
    }
    else if (!valueInRange(vp->df3.bpv, tmp)) {
      Tcl_SetErrorCode(itp, "oops, out-of-range value '",
              Tcl_GetString(av[i + 1]), "'.", NULL);
      return TCL_ERROR;
    }
    df3Error const ec = df3_put_voxel(&vp->df3, at, (uint32_t)tmp);
    if (ecOk != ec) {
      Tcl_SetErrorCode(itp, "oops, error writing DF3: '",
              df3_error_message(ec), "'.", NULL);
      return TCL_ERROR;
    }
  }

  Tcl_ResetResult(itp);
  return TCL_OK;
}

static int tokenCmdRget(Tcl_Interp *itp,
                        Volume *vp,
                        int ac AUU,
                        Tcl_Obj * const av[]) {
  df3Coord from, to;
  from = to = (df3Coord){0, 0, 0};
  if (TCL_OK != list2coord(itp, &from, av[0])) {
    Tcl_SetErrorCode(itp, "oops, bad LLC coordinate '",
            Tcl_GetString(av[0]), "'.", NULL);
    return TCL_ERROR;
  }
  else if (TCL_OK != list2coord(itp, &to, av[1])) {
    Tcl_SetErrorCode(itp, "oops, bad TRC coordinate '",
            Tcl_GetString(av[1]), "'.", NULL);
    return TCL_ERROR;
  }
  else if (!cornerCheck(itp, vp->df3.axes, from, to))
    return TCL_ERROR;

  Tcl_Obj *list = Tcl_NewListObj(0, NULL);
  Tcl_IncrRefCount(list);
  for (size_t z = from.z; z != (to.z + 1); ++z)
    for (size_t y = from.y; y != (to.y + 1); ++y)
      for (size_t x = from.x; x != (to.x + 1); ++x) {
        df3Coord const at = {x, y, z};
        uint32_t v = 0;
        df3Error const ec = df3_get_voxel(&vp->df3, at, &v);
        if (ecOk != ec) {
          if (app_diag)
            fprintf(stderr, "[II] df3 read error: '%s'.\n",
                    df3_handle_error(&vp->df3, false));
          Tcl_SetErrorCode(itp, "oops, df3 read error: '",
                  df3_error_message(ec), "'.", NULL);
          return TCL_ERROR;
        }
        Tcl_Obj *tmp = coord2list(at);
        Tcl_ListObjAppendElement(itp, list, tmp);
        Tcl_DecrRefCount(tmp);
        Tcl_ListObjAppendElement(itp, list, Tcl_NewLongObj((long)v));
      }

  Tcl_SetObjResult(itp, list);
  Tcl_DecrRefCount(list);
  return TCL_OK;
}


static int tokenCmdRput(Tcl_Interp *itp,
                        Volume *vp,
                        int ac AUU,
                        Tcl_Obj * const av[]) {
  if (vp->df3.readonly) {
    Tcl_SetErrorCode(itp, "oops, cannot modify 'readonly' volume.", NULL);
    return TCL_ERROR;
  }

  df3Coord from, to;
  from = to = (df3Coord){0, 0, 0};
  if (TCL_OK != list2coord(itp, &from, av[0])) {
    Tcl_SetErrorCode(itp, "oops, bad LLC coordinate '",
            Tcl_GetString(av[0]), "'.", NULL);
    return TCL_ERROR;
  }
  else if (TCL_OK != list2coord(itp, &to, av[1])) {
    Tcl_SetErrorCode(itp, "oops, bad TRC coordinate '",
            Tcl_GetString(av[1]), "'.", NULL);
    return TCL_ERROR;
  }
  else if (!cornerCheck(itp, vp->df3.axes, from, to))
    return TCL_ERROR;

  long val = 0;
  if (TCL_OK != Tcl_GetLongFromObj(itp, av[2], &val)) {
    Tcl_SetErrorCode(itp, "oops, expect (numeric) voxel value, got '",
            Tcl_GetString(av[2]), "'.", NULL);
    return TCL_ERROR;
  }
  else if (!valueInRange(vp->df3.bpv, val)) {
    Tcl_SetErrorCode(itp, "oops, out-of-range value '",
            Tcl_GetString(av[2]), "'.", NULL);
    return TCL_ERROR;
  }

  uint32_t const v = (uint32_t)val;
  for (size_t z = from.z; z != (to.z + 1); ++z)
    for (size_t y = from.y; y != (to.y + 1); ++y)
      for (size_t x = from.x; x != (to.x + 1); ++x) {
        df3Coord const at = {x, y, z};
        df3Error const ec = df3_put_voxel(&vp->df3, at, v);
        if (ecOk != ec) {
          Tcl_SetErrorCode(itp, "oops, error writing DF3: '",
                  df3_error_message(ec), "'.", NULL);
          return TCL_ERROR;
        }
      }

  Tcl_ResetResult(itp);
  return TCL_OK;
}


/* df3ObjCmd()
 * -------------------------------------------------------------------------- *
 * token command handler.
 */

static int df3ObjCmd(ClientData data,
                     Tcl_Interp *itp,
                     int ac,
                     Tcl_Obj * const av[]) {
  static const char * const scmd[] = {
    "axes", "bpv", "foreach", "get", "info", "put", "rget", "rput", NULL
  };
  Volume * const vp = (Volume * const)data;
  int ix;

  if (2 > ac) {
    Tcl_WrongNumArgs(itp, 1, av, oemesg);
    return TCL_ERROR;
  }
  else if (TCL_OK != Tcl_GetIndexFromObj(itp, av[1], scmd, "option", 0, &ix))
    return TCL_ERROR;

  if (2 == ac  &&  0 == ix)
    return tokenCmdAxes(itp, vp);
  else if (2 == ac  &&  1 == ix)
    return tokenCmdBpv(itp, vp);
  else if (2 == ac  &&  4 == ix)
    return tokenCmdInfo(itp, vp);
  else if (3 == ac  &&  2 == ix)
    return tokenCmdForeach(itp, vp, av[2]);
  else if (3 <= ac  &&  3 == ix)
    return tokenCmdGet(itp, vp, ac - 2, av + 2);
  else if (4 <= ac  &&  !(ac % 2)  &&  5 == ix)
    return tokenCmdPut(itp, vp, ac - 2, av + 2);
  else if (4 == ac  &&  6 == ix)
    return tokenCmdRget(itp, vp, ac - 2, av + 2);
  else if (5 == ac  &&  7 == ix)
    return tokenCmdRput(itp, vp, ac - 2, av + 2);

  Tcl_WrongNumArgs(itp, 1, av, oemesg);
  return TCL_ERROR;
}

/* df3CmdCreate()
 * -------------------------------------------------------------------------- *
 * leaves new volume "object" token in interpreter result.
 */

static int df3CmdCreate(Tcl_Interp *itp,
                        Df3tclData *dp,
                        int ac,
                        Tcl_Obj * const av[]) {

  df3Coord at = {0,0,0};
  size_t bpv = 0;
  bool rdonly = true;
  char const * const fnarg = Tcl_GetString(av[ac - 1]);

  for (size_t i = 0; i != (size_t const)(ac - 1); i += 2) {
    char const * const o = Tcl_GetString(av[i]);
    if (Tcl_StringMatch(o, "-a*")) {
      int const ec = list2coord(itp, &at, av[i + 1]);
      if (TCL_OK != ec) {
        Tcl_SetErrorCode(itp, "oops, error in 'axes'", NULL);
        return TCL_ERROR;
      }
      else if (!df3_check_axes(at)) {
        Tcl_SetErrorCode(itp, "oops, illegal 'axes' value(s)", NULL);
        return TCL_ERROR;
      }
    }
    else if (Tcl_StringMatch(o, "-b*")) {
      int tmp = 0;
      if (TCL_OK != Tcl_GetIntFromObj(itp, av[i + 1], &tmp)) {
        Tcl_SetErrorCode(itp, "oops, bad value '",
                Tcl_GetString(av[i + 1]), "'.", NULL);
        return TCL_ERROR;
      }
      else if (!df3_check_bpv((size_t const)tmp)) {
        Tcl_SetErrorCode(itp, "oops, illegal 'bpv' value", NULL);
        return TCL_ERROR;
      }
      bpv = (size_t)tmp;
    }
    else if (Tcl_StringMatch(o, "-r*")) {
      int tmp = 0;
      int const ec = Tcl_GetBooleanFromObj(itp, av[i + 1], &tmp);
      if (TCL_OK != ec) return TCL_ERROR;
      rdonly = tmp;
    }
    else {
      Tcl_SetErrorCode(itp, "oops, unknown option in 'create'", NULL);
      return TCL_ERROR;
    }
  }

  Volume *vp = (Volume *)Tcl_Alloc(sizeof(Volume));
  memset(vp, 0, sizeof(Volume));
  vp->df3 = DF3_INITHANDLE;
  vp->id = dp->nextid;
  dp->nextid += 1;

  /* open existing file, or, if have axes + bpv, create new one */
  if (4 > ac) {
    if (!dp->writemode && !rdonly) {
      Tcl_SetErrorCode(itp, "oops, no '-writemode' permission", NULL);
      goto label_fail;
    }
    bool flag = true;
    vp->df3.fname = storeFname(fnarg, DF3_EXTENSION, &flag);
    if (!flag) {
      Tcl_SetErrorCode(itp, "oops, can't use file name '", fnarg, "'", NULL);
      goto label_fail;
    }
    else if (!vp->df3.fname)
      APPPANIC;
    int const mode = rdonly ? R_OK : R_OK|W_OK;
    if (filePermCheck(vp->df3.fname, mode)) {
      Tcl_SetErrorCode(itp, "oops, no file (or perms) for '",
              vp->df3.fname, "'", NULL);
      goto label_fail;
    }
    df3Error const ec = volumeStartOpen(vp, rdonly);
    if (ecOk != ec) {
      Tcl_SetErrorCode(itp, "oops, opening file '",
              vp->df3.fname, "' failed", NULL);
      goto label_fail;
    }
  }
  else if (!(bpv  &&  at.x)) {
    Tcl_SetErrorCode(itp, "oops, need both '-axes' and '-bpv' "
                          "options to create volume.", NULL);
    goto label_fail;
  }
  else {
    char buf[PATH_MAX] = {CHRNUL};
    int n = PATH_MAX + 1;
    if (strrchr(fnarg, CHRPTS))
      n = snprintf(buf, PATH_MAX, "%s", fnarg);
    else
      n = snprintf(buf, PATH_MAX, "%s%c%s", dp->workdir, CHRPTS, fnarg);
    if (PATH_MAX <= n) APPPANIC;
    bool flag = true;
    vp->df3.fname = storeFname(buf, DF3_EXTENSION, &flag);
    if (!flag) {
      Tcl_SetErrorCode(itp, "oops, can't use file name '", buf, "'", NULL);
      goto label_fail;
    }
    else if (!vp->df3.fname)
      APPPANIC;
    vp->df3.axes = at;
    vp->df3.bpv = bpv;
    if (!filePermCheck(vp->df3.fname, R_OK)) {
      Tcl_SetErrorCode(itp, "oops, file '", vp->df3.fname, "' exists", NULL);
      goto label_fail;
    }
    df3Error const ec = volumeStartCreate(vp);
    if (ecOk != ec) {
      Tcl_SetErrorCode(itp, "oops, creating file '",
              vp->df3.fname, "' failed", NULL);
      goto label_fail;
    }
  }

  Tcl_Obj *token = Tcl_ObjPrintf(TOKEN ".%lu", vp->id);
  Tcl_IncrRefCount(token);

  int new = 0;
  Tcl_HashEntry *hp =
          Tcl_CreateHashEntry(&dp->hash, Tcl_GetString(token), &new);
  if (1 != new) APPPANIC;
  Tcl_SetHashValue(hp, (ClientData)vp);
  /* every token has a linked NS variable */
  char lnk[40] = {CHRNUL};
  sprintf(lnk, nslvmask, vp->id);
  if (TCL_OK != Tcl_LinkVar(itp, lnk, (void *)&vp->lv, TCL_LINK_UINT))
    goto label_fail;
  Tcl_CreateObjCommand(itp,
          Tcl_GetString(token), df3ObjCmd, (ClientData)vp, NULL);
  Tcl_SetObjResult(itp, token);
  Tcl_DecrRefCount(token);
  return TCL_OK;

label_fail:
  volumeFinish(itp, vp);
  return TCL_ERROR;
}

/* df3CmdDestroy()
 * -------------------------------------------------------------------------- *
 * remove the named volume object(s) from memory.
 */

static int df3CmdDestroy(Tcl_Interp *itp,
                         Df3tclData *dp,
                         int ac,
                         Tcl_Obj * const av[]) {
  Tcl_HashEntry *hp;
  Volume *vp;

  size_t i = 0;
  while (i < (size_t const)ac) {
    hp = Tcl_FindHashEntry(&dp->hash, Tcl_GetString(av[i]));
    if (NULL == hp) {
      Tcl_SetErrorCode(itp, "oops, volume \"", Tcl_GetString(av[i]),
              "\" not found.", NULL);
      return TCL_ERROR;
    }
    vp = Tcl_GetHashValue(hp);
    Tcl_DeleteHashEntry(hp);
    volumeFinish(itp, vp);
    i += 1;
  }

  Tcl_ResetResult(itp);
  return TCL_OK;
}

/* df3CmdNames()
 * -------------------------------------------------------------------------- *
 * return the list of (open) volumes.
 */

static int df3CmdNames(Tcl_Interp *itp, Df3tclData *dp) {
  char *k;
  Tcl_HashEntry *hp;
  Tcl_HashSearch search;
  Tcl_Obj *list;

  list = Tcl_NewListObj(0, NULL);
  Tcl_IncrRefCount(list);

  hp = Tcl_FirstHashEntry(&dp->hash, &search);
  while (NULL != hp) {
    k = Tcl_GetHashKey(&dp->hash, hp);
    if (TCL_OK != Tcl_ListObjAppendElement(itp, list, Tcl_NewStringObj(k, -1)))
      return TCL_ERROR;
    hp = Tcl_NextHashEntry(&search);
  }

  Tcl_SetObjResult(itp, list);
  Tcl_DecrRefCount(list);
  return TCL_OK;
}

/* df3Cmd()
 * -------------------------------------------------------------------------- *
 * the "package" command handler.  partially vet arguments, and return via
 * respective sub-command.
 */

static int df3Cmd(ClientData data,
                  Tcl_Interp *itp,
                  int ac,
                  Tcl_Obj * const av[]) {
  static const char * const scmd[] = {"create", "destroy", "names", NULL};
  Df3tclData *dp = (Df3tclData *)data;
  int ix;

  if (2 > ac) {
    Tcl_WrongNumArgs(itp, 1, av, oemesg);
    return TCL_ERROR;
  }
  else if (TCL_OK != Tcl_GetIndexFromObj(itp, av[1], scmd, "option", 0, &ix))
    return TCL_ERROR;

  /* after names and destroy have been dealt with, ix must point to
   * create, and the number of arguments must be odd and max total 9.
   */
  if (2 == ac  &&  2 == ix)
    return df3CmdNames(itp, dp);
  else if (3 <= ac  &&  1 == ix)
    return df3CmdDestroy(itp, dp, ac - 2, av + 2);
  else if (ix  ||  9 < ac  ||  !(ac % 2)) {
    Tcl_WrongNumArgs(itp, 1, av, oemesg);
    return TCL_ERROR;
  }

  return df3CmdCreate(itp, dp, ac - 2, av + 2);
}

/* prepTclif()
 * -------------------------------------------------------------------------- *
 * open data files.  tclsh-like environment.
 */

static void df3CmdEndUse(ClientData data) {
  Df3tclData *dp = (Df3tclData *)data;
  if (app_diag) {
    char *tmp = Tcl_HashStats(&dp->hash);
    fprintf(stderr, "[II] hash stats: %s.\n", tmp);
    Tcl_Free(tmp);
  }
  Tcl_HashSearch search;
  Tcl_HashEntry *hp = Tcl_FirstHashEntry(&dp->hash,  &search);
  while (hp) {
    Volume *vp = Tcl_GetHashValue(hp);
    Tcl_DeleteHashEntry(hp);
    volumeFinish(dp->interp, vp);
    hp = Tcl_FirstHashEntry(&dp->hash,  &search);
  }
  Tcl_DeleteHashTable(&dp->hash);
  return;
}

void prepTclif(Df3tclData *data, Tcl_Interp *itp, char **df) {
  Tcl_InitHashTable(&data->hash, TCL_STRING_KEYS);
  Tcl_CreateObjCommand(itp, TOKEN, df3Cmd, (ClientData)data, df3CmdEndUse);
  if (TCL_OK != Tcl_Eval(itp, nscreate)) {
    fprintf(stderr, "%s: oops, 'namespace eval' failed.\n", app_name);
    Tcl_Exit(EXIT_FAILURE);
  }
  /* "bad" hack, need interpreter in 'df3CmdEndUse' */
  data->interp = itp;

  int const mode = data->writemode ? (R_OK|W_OK) : R_OK;
  if (df)
    for (size_t i = 0; df[i]; ++i) {
      bool flag = true;
      char *tmp = storeFname(df[i], DF3_EXTENSION, &flag);
      if (!flag) {
        fprintf(stderr, "%s: oops, can't use file name '%s'.\n",
                app_name, df[i]);
        Tcl_Exit(EXIT_FAILURE);
      }
      else if (!tmp)
        APPPANIC;
      if (filePermCheck(tmp, mode)) {
        fprintf(stderr, "%s: oops, no file (or perms) for '%s'.\n",
                app_name, tmp);
        Tcl_Exit(EXIT_FAILURE);
      }
      free(tmp);
      /* large enough.  :-) */
      char s[PATH_MAX] = {CHRNUL};
      snprintf(s, PATH_MAX, smask, (int)!data->writemode , df[i]);
      if (TCL_OK != Tcl_Eval(itp, s)) {
        fprintf(stderr, "%s: oops, internal script failed.\n", app_name);
        Tcl_Exit(EXIT_FAILURE);
      }
    }

  return;
}

